#pragma once

#ifndef OPENCL_EXCEPTION_H
#define OPENCL_EXCEPTION_H


/**
	\file
	\brief Contains OpenCL::Exception class.
*/




#include "settings.h"




#if ( SETTING_OPENCL != 0 )




#include <CL/cl.h>

#include <string>
#include <stdexcept>




/*==============================================================================
	THROW_OPENCL_EXCEPTION macros
==============================================================================*/


/// \brief Throws an OpenCL::Exception labeled with the file, line, string \p what and OpenCL error code \p hResult.
#define THROW_OPENCL_EXCEPTION( what, result ) \
	throw ::OpenCL::Exception( __FILE__, __LINE__, ( what ), ( result ) )


/// \brief Evaluates \p what, and if it returns an OpenCL failure code, throws an OpenCL::Exception.
#define CHECK_OPENCL( what ) { \
	cl_int result = ( what ); \
	if ( result != CL_SUCCESS ) \
		THROW_OPENCL_EXCEPTION( #what, result ); \
}




namespace OpenCL {




/*==============================================================================
	Exception class
==============================================================================*/


/**
	\brief OpenCL exception class.

	OpenCL Exceptions should not be constructed explicitly. Instead, use the
	THROW_OPENCL_EXCEPTION() macro, which will construct and throw an exception
	annotated with the filename and line number of the failure.

	The difference between this class and the "default" exception class is that
	an OpenCL exception converts OpenCL error codes into error strings.

	The error message returned by what() has the following components:

	1. Filename and line number at which the exception was thrown.
	2. Description string \p what passed to the constructor.
	3. Description of the OpenCL error code passed to the constructor.
*/
struct Exception : public std::exception {

	Exception( char const* const file, unsigned int const line, std::string const& what, cl_int const result );


	virtual ~Exception() throw();


	virtual const char* what() const throw();    ///< Returns text (UTF-8) error message for this exception.


private:

	std::string m_what;
};




}    // namespace OpenCL




#endif    // SETTING_OPENCL




#endif    // OPENCL_EXCEPTION_H
