/**
	\file
	\brief Contains OpenCL::State methods.
*/




#include "settings.h"




#if ( SETTING_OPENCL != 0 )




#include "opencl_state.h"
#include "opencl_exception.h"
#include "exception.h"

#include <iostream>
#include <memory>




namespace OpenCL {




/*==============================================================================
	State methods
==============================================================================*/


State::State() : m_deviceID( NULL ) {

	// **TODO: use a cleaner form of logging than std::cout

	std::cout << "Searching for OpenCL device ...." << std::endl;
	{	cl_uint platformIDsLength = 0;
		clGetPlatformIDs( 0, NULL, &platformIDsLength );
		if ( platformIDsLength > 0 ) {

			std::unique_ptr< cl_platform_id[] > platformIDs( new cl_platform_id[ platformIDsLength ] );
			clGetPlatformIDs( platformIDsLength, platformIDs.get(), NULL );
			for ( unsigned int ii = 0; ii < platformIDsLength; ++ii ) {

				cl_uint deviceIDsLength = 0;
				clGetDeviceIDs( platformIDs[ ii ], CL_DEVICE_TYPE_GPU, 0, NULL, &deviceIDsLength );
				if ( deviceIDsLength > 0 ) {

					std::unique_ptr< cl_device_id[] > deviceIDs( new cl_device_id[ deviceIDsLength ] );
					clGetDeviceIDs( platformIDs[ ii ], CL_DEVICE_TYPE_GPU, deviceIDsLength, deviceIDs.get(), NULL );
					for ( unsigned int jj = 0; jj < deviceIDsLength; ++jj ) {

						size_t nameSize = 0;
						CHECK_OPENCL( clGetDeviceInfo( deviceIDs[ jj ], CL_DEVICE_NAME, 0, NULL, &nameSize ) );
						std::unique_ptr< cl_char[] > name( new cl_char[ nameSize ] );
						CHECK_OPENCL( clGetDeviceInfo( deviceIDs[ jj ], CL_DEVICE_NAME, nameSize, name.get(), NULL ) );

						cl_bool available = CL_TRUE;
						CHECK_OPENCL( clGetDeviceInfo( deviceIDs[ jj ], CL_DEVICE_AVAILABLE, sizeof( available ), &available, NULL ) );
						if ( ! available ) {

							std::cout << "\tCannot use \"" << name.get() << "\" because it is unavailable." << std::endl;
							continue;
						}

						cl_bool compilerAvailable = CL_TRUE;
						CHECK_OPENCL( clGetDeviceInfo( deviceIDs[ jj ], CL_DEVICE_COMPILER_AVAILABLE, sizeof( compilerAvailable ), &compilerAvailable, NULL ) );
						if ( ! compilerAvailable ) {

							std::cout << "\tCannot use \"" << name.get() << "\" because no compiler is available." << std::endl;
							continue;
						}

						cl_bool littleEndian = CL_FALSE;
						CHECK_OPENCL( clGetDeviceInfo( deviceIDs[ jj ], CL_DEVICE_ENDIAN_LITTLE, sizeof( littleEndian ), &littleEndian, NULL ) );
						if ( ! littleEndian ) {

							std::cout << "\tCannot use \"" << name.get() << "\" because it is big-endian." << std::endl;
							continue;
						}

						cl_bool imageSupport = CL_TRUE;
						CHECK_OPENCL( clGetDeviceInfo( deviceIDs[ jj ], CL_DEVICE_IMAGE_SUPPORT, sizeof( imageSupport ), &imageSupport, NULL ) );
						if ( ! imageSupport ) {

							std::cout << "\tCannot use \"" << name.get() << "\" because it does not support images." << std::endl;
							continue;
						}

						if ( m_deviceID != NULL ) {

							std::cout << "\tNot using \"" << name.get() << "\" because we have already chosen a device." << std::endl;
							continue;
						}

						std::cout << "\tUsing \"" << name.get() << "\"." << std::endl;
						m_deviceID = deviceIDs[ jj ];
					}
				}
			}
		}
	}
	if ( m_deviceID == NULL )
		THROW_EXCEPTION( "No usable GPU device found!" );

	cl_int result = CL_SUCCESS;

	m_context.reset( clCreateContext( NULL, 1, &m_deviceID, NULL, NULL, &result ) );
	if ( result != CL_SUCCESS )
		THROW_OPENCL_EXCEPTION( "clCreateContext failed!", result );

	m_queue.reset( clCreateCommandQueue( m_context.get(), m_deviceID, 0, &result ) );
	if ( result != CL_SUCCESS )
		THROW_OPENCL_EXCEPTION( "clCreateCommandQueue failed!", result );

	std::cout << "3D image formats supported by OpenCL device:" << std::endl;
	{	cl_uint imageFormatsLength = 0;
		clGetSupportedImageFormats( m_context.get(), CL_MEM_READ_ONLY, CL_MEM_OBJECT_IMAGE3D, 0, NULL, &imageFormatsLength );
		if ( imageFormatsLength > 0 ) {

			std::unique_ptr< cl_image_format[] > imageFormats( new cl_image_format[ imageFormatsLength ] );
			clGetSupportedImageFormats( m_context.get(), CL_MEM_READ_ONLY, CL_MEM_OBJECT_IMAGE3D, imageFormatsLength, imageFormats.get(), NULL );
			for ( unsigned int ii = 0; ii < imageFormatsLength; ++ii ) {

				std::cout << '\t';
				switch( imageFormats[ ii ].image_channel_order ) {
					case CL_R:         { std::cout << "CL_R";         break; }
					case CL_A:         { std::cout << "CL_A";         break; }
					case CL_INTENSITY: { std::cout << "CL_INTENSITY"; break; }
					case CL_LUMINANCE: { std::cout << "CL_LUMINANCE"; break; }
					case CL_RG:        { std::cout << "CL_RG";        break; }
					case CL_RA:        { std::cout << "CL_RA";        break; }
					case CL_RGB:       { std::cout << "CL_RGB";       break; }
					case CL_RGBA:      { std::cout << "CL_RGBA";      break; }
					case CL_ARGB:      { std::cout << "CL_ARGB";      break; }
					case CL_BGRA:      { std::cout << "CL_BGRA";      break; }
					default: { std::cout << "<unknown channel order>"; break; }
				}

				std::cout << ' ';
				switch( imageFormats[ ii ].image_channel_data_type ) {
					case CL_SNORM_INT8:       { std::cout << "CL_SNORM_INT8";       break; }
					case CL_SNORM_INT16:      { std::cout << "CL_SNORM_INT16";      break; }
					case CL_UNORM_INT8:       { std::cout << "CL_UNORM_INT8";       break; }
					case CL_UNORM_INT16:      { std::cout << "CL_UNORM_INT16";      break; }
					case CL_UNORM_SHORT_565:  { std::cout << "CL_UNORM_SHORT_565";  break; }
					case CL_UNORM_SHORT_555:  { std::cout << "CL_UNORM_SHORT_555";  break; }
					case CL_UNORM_INT_101010: { std::cout << "CL_UNORM_INT_101010"; break; }
					case CL_SIGNED_INT8:      { std::cout << "CL_SIGNED_INT8";      break; }
					case CL_SIGNED_INT16:     { std::cout << "CL_SIGNED_INT16";     break; }
					case CL_SIGNED_INT32:     { std::cout << "CL_SIGNED_INT32";     break; }
					case CL_UNSIGNED_INT8:    { std::cout << "CL_UNSIGNED_INT8";    break; }
					case CL_UNSIGNED_INT16:   { std::cout << "CL_UNSIGNED_INT16";   break; }
					case CL_UNSIGNED_INT32:   { std::cout << "CL_UNSIGNED_INT32";   break; }
					case CL_HALF_FLOAT:       { std::cout << "CL_HALF_FLOAT";       break; }
					case CL_FLOAT:            { std::cout << "CL_FLOAT";            break; }
					default: { std::cout << "<unknown channel data type>"; break; }
				}
				std::cout << std::endl;
			}
		}
	}
}


State::~State() {

	if ( m_queue ) {

		clFlush(  m_queue.get() );
		clFinish( m_queue.get() );
	}
}




}    // namespace OpenCL




#endif    // SETTING_OPENCL
