#pragma once

#ifndef OPENCL_STATE_H
#define OPENCL_STATE_H


/**
	\file
	\brief Contains OpenCL::State class.
*/




#include "settings.h"




#if ( SETTING_OPENCL != 0 )




#include "opencl_deleter.h"

#include <CL/cl.h>

#include <memory>




namespace OpenCL {




/*==============================================================================
	State class
==============================================================================*/


/**
	\brief OpenCL state class.

	The constructor attempts to create the following three objects:

	1. An OpenCL device identifier (for the first GPU found in the system).
	2. An OpenCL context (for this device).
	3. An OpenCL command queue (also for this device).

	If any of the creation calls fail, than an OpenCL Exception will be thrown.
	The three contained objects can be accessed via the \c DeviceID, \c Context
	and \c Queue accessor functions, respectively.

	\todo Check endian-ness.
	\todo Check supported image formats (in particular, RGBA32 and RGB16, the two supported in color.h).
*/
struct State {

	State();

	~State();


	inline cl_device_id     DeviceID() const;    ///< \brief Returns OpenCL device identifier of first GPU in the system.
	inline cl_context       Context()  const;    ///< \brief Returns OpenCL context for the device returned by DeviceID().
	inline cl_command_queue Queue()    const;    ///< \brief Returns OpenCL command queue for the device returned by DeviceID().


private:

	cl_device_id m_deviceID;
	std::unique_ptr< _cl_context,       Deleter< _cl_context        > > m_context;
	std::unique_ptr< _cl_command_queue, Deleter< _cl_command_queue  > > m_queue;
};




/*==============================================================================
	State methods
==============================================================================*/


cl_device_id State::DeviceID() const {

	return m_deviceID;
}


cl_context State::Context() const {

	return m_context.get();
}


cl_command_queue State::Queue() const {

	return m_queue.get();
}




}    // namespace OpenCL




#endif    // SETTING_OPENCL




#endif    // OPENCL_STATE_H
