#pragma once

#ifndef RENDERER_DETAIL_TRACE_H
#define RENDERER_DETAIL_TRACE_H




#include "synchronized.h"
#include "octree.h"
#include "vector.h"
#include "helpers.h"
#include "settings.h"

#include <boost/math/special_functions.hpp>    // MSVC does not yet support std::isfinite

#include <algorithm>
#include <limits>
#include <cmath>

#include <assert.h>
#include <stdint.h>




namespace Renderer {


namespace detail {




/*==============================================================================
	Function declarations
==============================================================================*/


double Project( Vector< double, 3 >& position, Vector< double, 3 > const& ray );




/*==============================================================================
	TraceTexture function
==============================================================================*/


template< int t_ColorFormat, int t_Platform >
int TraceTexture(
	uint8_t* const colorDestination,
	float* const depthDestination,
	typename Synchronized::TextureHeap< t_ColorFormat, t_Platform >::Pointer const& texture,
	double lambda,
	Vector< double, 3 > remaining,
	Vector< double, 3 > remainingMaximum,
	unsigned int const origin
)
{
	static unsigned int const maskTable[] = { 0, 3, 12, 15, 48, 51, 60, 63 };    // abc -> aabbcc (binary)

	unsigned int const mask = maskTable[ origin ];
	unsigned int index = 0;

	remainingMaximum /= 2;
	if ( remaining[ 0 ] >= remainingMaximum[ 0 ] ) {

		index += 2;
		remaining[ 0 ] -= remainingMaximum[ 0 ];
	}
	if ( remaining[ 1 ] >= remainingMaximum[ 1 ] ) {

		index += 8;
		remaining[ 1 ] -= remainingMaximum[ 1 ];
	}
	if ( remaining[ 2 ] >= remainingMaximum[ 2 ] ) {

		index += 32;
		remaining[ 2 ] -= remainingMaximum[ 2 ];
	}

	remainingMaximum /= 2;
	if ( remaining[ 0 ] >= remainingMaximum[ 0 ] ) {

		index += 1;
		remaining[ 0 ] -= remainingMaximum[ 0 ];
	}
	if ( remaining[ 1 ] >= remainingMaximum[ 1 ] ) {

		index += 4;
		remaining[ 1 ] -= remainingMaximum[ 1 ];
	}
	if ( remaining[ 2 ] >= remainingMaximum[ 2 ] ) {

		index += 16;
		remaining[ 2 ] -= remainingMaximum[ 2 ];
	}

	int comparisons = -1;
	while ( comparisons < 0 ) {

		assert( ( index ^ mask ) < 64 );
		Color< t_ColorFormat > const& color = texture[ index ^ mask ];
		if ( color.Packed() != 0 ) {    // **NOTE: black is transparent

			// **NOTE: BGRA
			colorDestination[ 0 ] = color.Blue();
			colorDestination[ 1 ] = color.Green();
			colorDestination[ 2 ] = color.Red();
			colorDestination[ 3 ] = 0xff;

			*depthDestination = lambda;

			comparisons = -1;
			break;    // **NOTE: break
		}

		comparisons = (
			( ( remaining[ 0 ] < remaining[ 1 ] ) ? 1 : 0 ) |
			( ( remaining[ 1 ] < remaining[ 2 ] ) ? 2 : 0 ) |
			( ( remaining[ 2 ] < remaining[ 0 ] ) ? 4 : 0 )
		);
		switch( comparisons ) {

			case 0: case 7:    // any are fine--fall through
			case 1: case 3: {

				if ( ( index & 3 ) != 0 ) {

					--index;

					lambda += remaining[ 0 ];
					remaining[ 1 ] -= remaining[ 0 ];
					remaining[ 2 ] -= remaining[ 0 ];
					remaining[ 0 ] = remainingMaximum[ 0 ];

					comparisons = -1;
				}

				break;
			}

			case 2: case 6: {

				if ( ( index & 12 ) != 0 ) {

					index -= 4;

					lambda += remaining[ 1 ];
					remaining[ 0 ] -= remaining[ 1 ];
					remaining[ 2 ] -= remaining[ 1 ];
					remaining[ 1 ] = remainingMaximum[ 1 ];

					comparisons = -1;
				}

				break;
			}

			case 4: case 5: {

				if ( ( index & 48 ) != 0 ) {

					index -= 16;

					lambda += remaining[ 2 ];
					remaining[ 0 ] -= remaining[ 2 ];
					remaining[ 1 ] -= remaining[ 2 ];
					remaining[ 2 ] = remainingMaximum[ 2 ];

					comparisons = -1;
				}

				break;
			}
		}
	}

	return comparisons;
}




/*==============================================================================
	Trace function
==============================================================================*/


template< int t_ColorFormat, int t_Platform >
void Trace(
	uint8_t* const colorDestination,
	float* const depthDestination,
	typename Synchronized::Heap< t_Platform >::template Pointer< Octree::Node< t_ColorFormat, t_Platform > > const& pRoot,
	double const spread,
	Vector< double, 3 > const& position,
	Vector< double, 3 > const& ray
)
{
	double const epsilon = std::numeric_limits< double >::epsilon();

	Vector< double, 3 > projected = position;
	double lambda = Project( projected, ray );
	if ( ! boost::math::isfinite( lambda ) )
		return;    // **NOTE: return

	unsigned int origin = 7;
	Vector< double, 3 > remaining = 1 - projected;
	Vector< double, 3 > delta = ray;
	if ( delta[ 0 ] < 0 ) {

		origin ^= 1;
		remaining[ 0 ] = projected[ 0 ];
		delta[ 0 ] = -delta[ 0 ];
	}
	if ( delta[ 1 ] < 0 ) {

		origin ^= 2;
		remaining[ 1 ] = projected[ 1 ];
		delta[ 1 ] = -delta[ 1 ];
	}
	if ( delta[ 2 ] < 0 ) {

		origin ^= 4;
		remaining[ 2 ] = projected[ 2 ];
		delta[ 2 ] = -delta[ 2 ];
	}

	// deltas must be strictly positive
	delta = std::max( delta, epsilon );

#if ( SETTING_MIPMAP != 0 )
	Vector< double, 3 > distance = std::abs( position - projected ) - ( 1 - remaining );

	double sideLength = 1;
	double distanceThresholds[ SETTING_TRACE_STACK_SIZE ];
	distanceThresholds[ 0 ] = 0.125 / spread;    // 0.5 because we want to compare to the size at the *next* depth, and 0.25 because we draw 4x4x4 textures
	for ( unsigned int ii = 1; ii < ARRAYLENGTH( distanceThresholds ); ++ii )
		distanceThresholds[ ii ] = distanceThresholds[ ii - 1 ] / 2;
#endif    // SETTING_MIPMAP

	Vector< double, 3 > remainingMaximum = 1 / delta;
	remaining *= remainingMaximum;

	Octree::Node< t_ColorFormat, t_Platform > const* childrenStack[ SETTING_TRACE_STACK_SIZE ];
	unsigned int indexStack[ SETTING_TRACE_STACK_SIZE ];
	unsigned int depth = 0;

	Octree::Node< t_ColorFormat, t_Platform > const* children = pRoot.get();
	unsigned int index = 0;
	do {

		// ==== descend until we reach a leaf node or the maximum mipmap level ====

		while ( children[ index ].children ) {

#if ( SETTING_MIPMAP != 0 )
			if ( ( distance[ 0 ] > distanceThresholds[ depth ] ) || ( distance[ 1 ] > distanceThresholds[ depth ] ) || ( distance[ 2 ] > distanceThresholds[ depth ] ) )
				break;
#endif    // SETTING_MIPMAP

			assert( depth < ARRAYLENGTH( childrenStack ) );
			assert( depth < ARRAYLENGTH( indexStack    ) );
			childrenStack[ depth ] = children;
			indexStack[    depth ] = index;
			++depth;

			children = children[ index ].children.get();
			index = origin;

#if ( SETTING_MIPMAP != 0 )
			sideLength /= 2;
#endif    // SETTING_MIPMAP
			remainingMaximum /= 2;

			if ( remaining[ 0 ] >= remainingMaximum[ 0 ] ) {

				remaining[ 0 ] -= remainingMaximum[ 0 ];
				index ^= 1;
			}
#if ( SETTING_MIPMAP != 0 )
			else
				distance[ 0 ] += sideLength;
#endif    // SETTING_MIPMAP

			if ( remaining[ 1 ] >= remainingMaximum[ 1 ] ) {

				remaining[ 1 ] -= remainingMaximum[ 1 ];
				index ^= 2;
			}
#if ( SETTING_MIPMAP != 0 )
			else
				distance[ 1 ] += sideLength;
#endif    // SETTING_MIPMAP

			if ( remaining[ 2 ] >= remainingMaximum[ 2 ] ) {

				remaining[ 2 ] -= remainingMaximum[ 2 ];
				index ^= 4;
			}
#if ( SETTING_MIPMAP != 0 )
			else
				distance[ 2 ] += sideLength;
#endif    // SETTING_MIPMAP

			assert( index < 8 );
		}

		// ==== try to draw the 4x4x4 texture stored at this node, if any ====

		int comparisons = -1;
		if ( children[ index ].texture ) {

			comparisons = TraceTexture< t_ColorFormat, t_Platform >(
				colorDestination,
				depthDestination,
				children[ index ].texture,
				lambda,
				remaining,
				remainingMaximum,
				origin
			);

			if ( comparisons < 0 )
				break;    // **NOTE: break
		}
		else {

			comparisons = (
				( ( remaining[ 0 ] < remaining[ 1 ] ) ? 1 : 0 ) |
				( ( remaining[ 1 ] < remaining[ 2 ] ) ? 2 : 0 ) |
				( ( remaining[ 2 ] < remaining[ 0 ] ) ? 4 : 0 )
			);
		}

		// ==== step to the next adjacent voxel, ascending as necessary ====

		switch( comparisons ) {

			case 0: case 7:    // any are fine--fall through
			case 1: case 3: {

#if ( SETTING_MIPMAP != 0 )
				distance[ 0 ] += sideLength;
#endif    // SETTING_MIPMAP

				lambda += remaining[ 0 ];
				remaining[ 1 ] -= remaining[ 0 ];
				remaining[ 2 ] -= remaining[ 0 ];

				do {

					unsigned int const indexOrigin = ( index ^ origin );
					if ( indexOrigin & 1 ) {

						index ^= 1;
						break;
					}
					else {

						if ( indexOrigin & 2 )
							remaining[ 1 ] += remainingMaximum[ 1 ];
#if ( SETTING_MIPMAP != 0 )
						else
							distance[ 1 ] -= sideLength;
#endif    // SETTING_MIPMAP

						if ( indexOrigin & 4 )
							remaining[ 2 ] += remainingMaximum[ 2 ];
#if ( SETTING_MIPMAP != 0 )
						else
							distance[ 2 ] -= sideLength;
#endif    // SETTING_MIPMAP

#if ( SETTING_MIPMAP != 0 )
						sideLength *= 2;
#endif    // SETTING_MIPMAP
						remainingMaximum *= 2;

						assert( depth > 0 );
						--depth;
						children = childrenStack[ depth ];
						index    = indexStack[    depth ];
					}

				} while ( depth > 0 );

				remaining[ 0 ] = remainingMaximum[ 0 ];

				break;
			}

			case 2: case 6: {

#if ( SETTING_MIPMAP != 0 )
				distance[ 1 ] += sideLength;
#endif    // SETTING_MIPMAP

				lambda += remaining[ 1 ];
				remaining[ 0 ] -= remaining[ 1 ];
				remaining[ 2 ] -= remaining[ 1 ];

				do {

					unsigned int const indexOrigin = ( index ^ origin );
					if ( indexOrigin & 2 ) {

						index ^= 2;
						break;
					}
					else {

						if ( indexOrigin & 1 )
							remaining[ 0 ] += remainingMaximum[ 0 ];
#if ( SETTING_MIPMAP != 0 )
						else
							distance[ 0 ] -= sideLength;
#endif    // SETTING_MIPMAP

						if ( indexOrigin & 4 )
							remaining[ 2 ] += remainingMaximum[ 2 ];
#if ( SETTING_MIPMAP != 0 )
						else
							distance[ 2 ] -= sideLength;
#endif    // SETTING_MIPMAP

#if ( SETTING_MIPMAP != 0 )
						sideLength *= 2;
#endif    // SETTING_MIPMAP
						remainingMaximum *= 2;

						assert( depth > 0 );
						--depth;
						children = childrenStack[ depth ];
						index    = indexStack[    depth ];
					}

				} while ( depth > 0 );

				remaining[ 1 ] = remainingMaximum[ 1 ];

				break;
			}

			case 4: case 5: {

#if ( SETTING_MIPMAP != 0 )
				distance[ 2 ] += sideLength;
#endif    // SETTING_MIPMAP

				lambda += remaining[ 2 ];
				remaining[ 0 ] -= remaining[ 2 ];
				remaining[ 1 ] -= remaining[ 2 ];

				do {

					unsigned int const indexOrigin = ( index ^ origin );
					if ( indexOrigin & 4 ) {

						index ^= 4 ;
						break;
					}
					else {

						if ( indexOrigin & 1 )
							remaining[ 0 ] += remainingMaximum[ 0 ];
#if ( SETTING_MIPMAP != 0 )
						else
							distance[ 0 ] -= sideLength;
#endif    // SETTING_MIPMAP

						if ( indexOrigin & 2 )
							remaining[ 1 ] += remainingMaximum[ 1 ];
#if ( SETTING_MIPMAP != 0 )
						else
							distance[ 1 ] -= sideLength;
#endif    // SETTING_MIPMAP

#if ( SETTING_MIPMAP != 0 )
						sideLength *= 2;
#endif    // SETTING_MIPMAP
						remainingMaximum *= 2;

						assert( depth > 0 );
						--depth;
						children = childrenStack[ depth ];
						index    = indexStack[    depth ];
					}

				} while ( depth > 0 );

				remaining[ 2 ] = remainingMaximum[ 2 ];

				break;
			}
		}

	} while ( depth > 0 );
}




}    // namespace detail


}    // namespace Renderer




#endif    // RENDERER_DETAIL_TRACE_H
