/**
	\file
	\brief Contains OpenCL code for raytracing a scene.
*/




/*==============================================================================
	Node structure
==============================================================================*/


struct Node {

	uint texture;
	uint children;
};




/*==============================================================================
	Morton Z-order functions
==============================================================================*/


/**
	\brief Maps 2D coordinates to Morton Z-order indices.

	\param[in]  coordinate  The 2D coordinates.
	\param[in]  dimension   Dimensions of the region to which the coordinates belong.
	\return  The Morton Z-order index.

	Suppose that the coordinates represent positions in a buffer of the given
	dimensions. This function will return the (row-major) Morton Z-order index
	corresponding to these coordinates. The \ref InverseMortonZOrder function will
	perform the reverse operation (i.e. it is the inverse of this function). The
	\p dimension argument does not need to be a power of two.

	For example, if \p dimension.{x,y}={9,6}, then coordinates will map to
	indices as follows:

	\f[
	   \begin{array}{ccccccccc}
	    0 &  1 &  4 &  5 & 16 & 17 & 20 & 21 & 48 \\
	    2 &  3 &  6 &  7 & 18 & 19 & 22 & 23 & 49 \\
	    8 &  9 & 12 & 13 & 24 & 25 & 28 & 29 & 50 \\
	   10 & 11 & 14 & 15 & 26 & 27 & 30 & 31 & 51 \\
	   32 & 33 & 36 & 37 & 40 & 41 & 44 & 45 & 52 \\
	   34 & 35 & 38 & 39 & 42 & 43 & 46 & 47 & 53
	   \end{array}
	\f]

	\see \ref InverseMortonZOrder
*/
uint MortonZOrder( uint2 coordinate, uint2 dimension ) {

	uint result = 0;

	int shift = 0;
	for ( ; any( ( 1u << shift ) < dimension ); ++shift );
	for ( --shift; shift >= 0; --shift ) {

		uint const bit = ( 1u << shift );

		if ( coordinate.y & bit ) {

			dimension.y -= bit;
			result += ( dimension.x << shift );
		}
		else if ( dimension.y > bit )
			dimension.y = bit;

		if ( coordinate.x & bit ) {

			dimension.x -= bit;
			result += ( dimension.y << shift );
		}
		else if ( dimension.x > bit )
			dimension.x = bit;
	}

	return result;
}


/**
	\brief Maps Morton Z-order indices to 2D coordinates.

	\param[in]  index      The Morton Z-order index.
	\param[in]  dimension  Dimensions of the region to which the coordinates belong.
	\return  The 2D coordinates.

	This function will return the coordinates which should be considered by the
	"index"th thread, if it wants to work in Morton Z-order, over a buffer of the
	given size. In other words, this is an inverse Z-order mapping (and this
	function is the inverse of \ref MortonZOrder). The \p dimension argument does
	not need to be a power of two.

	\see \ref MortonZOrder
*/
uint2 InverseMortonZOrder( uint index, uint2 dimension ) {

	uint2 result = ( uint2 )( 0, 0 );

	int shift = 0;
	for ( ; any( ( 1u << shift ) < dimension ); ++shift );
	for ( --shift; shift >= 0; --shift ) {

		uint const bit = ( 1u << shift );

		{	uint const shifted = ( dimension.x << shift );
			if ( index >= shifted ) {

				result.y += bit;
				dimension.y -= bit;
				index -= shifted;
			}
			else if ( dimension.y > bit )
				dimension.y = bit;
		}

		{	uint const shifted = ( dimension.y << shift );
			if ( index >= shifted ) {

				result.x += bit;
				dimension.x -= bit;
				index -= shifted;
			}
			else if ( dimension.x > bit )
				dimension.x = bit;
		}
	}

	return result;
}




/*==============================================================================
	Project function
==============================================================================*/


/**
	\brief Projects a ray onto the unit cube.

	\param[in,out]  pProjected  Origin of the ray on input, intersection with unit cube on output.
	\param[in]      ray         Direction in which the ray propagates.
	\return  True if the ray intersects the unit cube, false otherwise.

	If the ray never intersects with the unit cube \f$[0,1]^3\f$, then this
	function will return false. Otherwise it will store the first intersection
	point in \p *pProjected, and return true. If the origin of the ray is
	inside the cube, then \p *pProjected will not be changed, but the function
	will return true.
*/
float Project( __private float3* pProjected, float3 const ray ) {

	if ( 
		( pProjected->x >= 0 ) && ( pProjected->x <= 1 ) &&
		( pProjected->y >= 0 ) && ( pProjected->y <= 1 ) &&
		( pProjected->z >= 0 ) && ( pProjected->z <= 1 )
	)
	{
		return 0;    // **NOTE: return
	}
	else {

		if ( pProjected->x < 0 ) {

			if ( ray.x > 0 ) {

				float const lambda = -pProjected->x / ray.x;
				float3 projected = *pProjected + lambda * ray;
				if (
					( projected.y >= 0 ) && ( projected.y <= 1 ) &&
					( projected.z >= 0 ) && ( projected.z <= 1 )
				)
				{
					projected.x = 0;    // just to make sure
					*pProjected = projected;
					return lambda;    // **NOTE: return
				}
			}
		}
		else if ( pProjected->x > 1 ) {

			if ( ray.x < 0 ) {

				float const lambda = ( 1 - pProjected->x ) / ray.x;
				float3 projected = *pProjected + lambda * ray;
				if (
					( projected.y >= 0 ) && ( projected.y <= 1 ) &&
					( projected.z >= 0 ) && ( projected.z <= 1 )
				)
				{
					projected.x = 1;    // just to make sure
					*pProjected = projected;
					return lambda;    // **NOTE: return
				}
			}
		}

		if ( pProjected->y < 0 ) {

			if ( ray.y > 0 ) {

				float const lambda = -pProjected->y / ray.y;
				float3 projected = *pProjected + lambda * ray;
				if (
					( projected.x >= 0 ) && ( projected.x <= 1 ) &&
					( projected.z >= 0 ) && ( projected.z <= 1 )
				)
				{
					projected.y = 0;    // just to make sure
					*pProjected = projected;
					return lambda;    // **NOTE: return
				}
			}
		}
		else if ( pProjected->y > 1 ) {

			if ( ray.y < 0 ) {

				float const lambda = ( 1 - pProjected->y ) / ray.y;
				float3 projected = *pProjected + lambda * ray;
				if (
					( projected.x >= 0 ) && ( projected.x <= 1 ) &&
					( projected.z >= 0 ) && ( projected.z <= 1 )
				)
				{
					projected.y = 1;    // just to make sure
					*pProjected = projected;
					return lambda;    // **NOTE: return
				}
			}
		}

		if ( pProjected->z < 0 ) {

			if ( ray.z > 0 ) {

				float const lambda = -pProjected->z / ray.z;
				float3 projected = *pProjected + lambda * ray;
				if (
					( projected.x >= 0 ) && ( projected.x <= 1 ) &&
					( projected.y >= 0 ) && ( projected.y <= 1 )
				)
				{
					projected.z = 0;    // just to make sure
					*pProjected = projected;
					return lambda;    // **NOTE: return
				}
			}
		}
		else if ( pProjected->z > 1 ) {

			if ( ray.z < 0 ) {

				float const lambda = ( 1 - pProjected->z ) / ray.z;
				float3 projected = *pProjected + lambda * ray;
				if (
					( projected.x >= 0 ) && ( projected.x <= 1 ) &&
					( projected.y >= 0 ) && ( projected.y <= 1 )
				)
				{
					projected.z = 1;    // just to make sure
					*pProjected = projected;
					return lambda;    // **NOTE: return
				}
			}
		}
	}

	return INFINITY;
}




/*==============================================================================
	TraceTexture function
==============================================================================*/


int TraceTexture(
	__private uint*  const pColor,
	__private float* const pLambda,
	__read_only image3d_t image,
	uint index,
	__private float3* const pRemaining,
	__private float3* const pRemainingMaximum,
	int const origin
)
{
	sampler_t const sampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_NONE | CLK_FILTER_NEAREST;

	uint const mask = ( ( origin & 4 ) ? ( 3 << 22 ) : 0 ) | ( ( origin & 2 ) ? ( 3 << 11 ) : 0 ) | ( ( origin & 1 ) ? 3 : 0 );

	*pRemainingMaximum /= 2;
	if ( pRemaining->x >= pRemainingMaximum->x ) {

		index += 2;
		pRemaining->x -= pRemainingMaximum->x;
	}
	if ( pRemaining->y >= pRemainingMaximum->y ) {

		index += ( 2 << 11 );
		pRemaining->y -= pRemainingMaximum->y;
	}
	if ( pRemaining->z >= pRemainingMaximum->z ) {

		index += ( 2 << 22 );
		pRemaining->z -= pRemainingMaximum->z;
	}

	*pRemainingMaximum /= 2;
	if ( pRemaining->x >= pRemainingMaximum->x ) {

		index += 1;
		pRemaining->x -= pRemainingMaximum->x;
	}
	if ( pRemaining->y >= pRemainingMaximum->y ) {

		index += ( 1u << 11 );
		pRemaining->y -= pRemainingMaximum->y;
	}
	if ( pRemaining->z >= pRemainingMaximum->z ) {

		index += ( 1u << 22 );
		pRemaining->z -= pRemainingMaximum->z;
	}

	int comparisons = 0;
	for ( ; ; ) {

		uint const source = ( index ^ mask );
		int4 coordinate = ( int4 )( source & 0x7ff, ( source >> 11 ) & 0x7ff, source >> 22, 0 );
		float4 const color = read_imagef(
			image,
			sampler,
			coordinate
		);
		if ( ( color.x + color.y + color.z ) > 0 ) {    // **NOTE: black is transparent

			comparisons = -1;
			*pColor = 0xff000000 + ( ( uint )( color.x * 255.99 ) << 16 ) + ( ( uint )( color.y * 255.99 ) << 8 ) + ( uint )( color.z * 255.99 );    // **NOTE: BGRA
			goto texture_loop_end_label;    // **NOTE: goto
		}

		comparisons = (
			( ( pRemaining->x < pRemaining->y ) ? 1 : 0 ) |
			( ( pRemaining->y < pRemaining->z ) ? 2 : 0 ) |
			( ( pRemaining->z < pRemaining->x ) ? 4 : 0 )
		);
		switch( comparisons ) {

			case 0: case 7:    // any are fine--fall through
			case 1: case 3: {

				*pLambda += pRemaining->x;
				pRemaining->y -= pRemaining->x;
				pRemaining->z -= pRemaining->x;
				pRemaining->x = pRemainingMaximum->x;

				if ( ( index & 3 ) == 0 ) {

					index |= 3;
					goto texture_loop_end_label;    // **NOTE: goto
				}

				index -= 1;

				break;
			}

			case 2: case 6: {

				*pLambda += pRemaining->y;
				pRemaining->x -= pRemaining->y;
				pRemaining->z -= pRemaining->y;
				pRemaining->y = pRemainingMaximum->y;

				if ( ( ( index >> 11 ) & 3 ) == 0 ) {

					index |= ( 3u << 11 );
					goto texture_loop_end_label;    // **NOTE: goto
				}

				index -= ( 1u << 11 );

				break;
			}

			case 4: case 5: {

				*pLambda += pRemaining->z;
				pRemaining->x -= pRemaining->z;
				pRemaining->y -= pRemaining->z;
				pRemaining->z = pRemainingMaximum->z;

				if ( ( ( index >> 22 ) & 3 ) == 0 ) {

					index |= ( 3u << 22 );
					goto texture_loop_end_label;    // **NOTE: goto
				}

				index -= ( 1u << 22 );

				break;
			}
		}
	}
texture_loop_end_label:;

	pRemaining->x += pRemainingMaximum->x * (   index         & 3 );
	pRemaining->y += pRemainingMaximum->y * ( ( index >> 11 ) & 3 );
	pRemaining->z += pRemainingMaximum->z * ( ( index >> 22 ) & 3 );

	*pRemainingMaximum *= 4;

	return comparisons;
}




/*==============================================================================
	Trace function
==============================================================================*/


void Trace(
	__global uint*  const colorDestination,
	__global float* const depthDestination,
	float const spread,
	__global void const* const heap,
	__read_only image3d_t image,
	uint const root,
	float3 const pp,
	float3 const ray
)
{
	float const epsilon = 1e-6;

	uint color = 0;
	float3 projected = pp;
	float lambda = Project( &projected, ray );

	if ( isfinite( lambda ) ) {

		int origin = 7;
		float3 remaining = 1 - projected;
		float3 delta = ray;
		if ( delta.x < 0 ) {

			origin ^= 1;
			remaining.x = projected.x;
			delta.x = -delta.x;
		}
		if ( delta.y < 0 ) {

			origin ^= 2;
			remaining.y = projected.y;
			delta.y = -delta.y;
		}
		if ( delta.z < 0 ) {

			origin ^= 4;
			remaining.z = projected.z;
			delta.z = -delta.z;
		}
		delta = max( delta, epsilon );

#if ( SETTING_MIPMAP != 0 )
		float3 const mipmapCoordinate = fabs( pp - projected ) + remaining;
		float const spreadInverse = 0.125 / spread;    // 0.5 because we want to compare to the size at the *next* depth, and 0.25 because we draw 4x4x4 textures
#endif    // SETTING_MIPMAP

		float3 remainingMaximum = 1 / delta;
		remaining *= remainingMaximum;

		uint3 coordinate = ( uint3 )( 0, 0, 0 );
		int coordinateDepth = 0;

		__global struct Node const* pNode = ( __global struct Node const* )( ( __global uchar const* )heap + root );

		__global struct Node const* childrenStack[ SETTING_TRACE_STACK_SIZE ];
		int depth = 0;

		for ( ; ; ) {

			// ==== align the tree depth with the coordinate depth ====

			// descend until we reach the coordinate depth or a leaf
			while ( ( depth < coordinateDepth ) && ( pNode->children != OPENCL_NULL ) ) {

#if ( SETTING_MIPMAP != 0 )
				uint3 const shiftedCoordinate = ( coordinate >> ( coordinateDepth - depth ) );
#if 0
				float3 const mipmapDistance = ldexp( mipmapCoordinate, depth ) - ( float3 )( shiftedCoordinate.x, shiftedCoordinate.y, shiftedCoordinate.z );
#else    // 0/1
				float const sideLengthInverse = ( 1 << depth );
				float3 const mipmapDistance = mipmapCoordinate * sideLengthInverse - ( float3 )( shiftedCoordinate.x, shiftedCoordinate.y, shiftedCoordinate.z );
#endif    // 0/1
				if ( any( mipmapDistance >= spreadInverse ) )
					break;    // **NOTE: break
#endif    // SETTING_MIPMAP

				pNode = ( __global struct Node const* )( ( __global uchar const* )heap + pNode->children );
				childrenStack[ depth ] = pNode;
				++depth;

				uint const mask = ( 1 << ( coordinateDepth - depth ) );
				pNode += origin ^ (
					( ( coordinate.x & mask ) ? 1 : 0 ) |
					( ( coordinate.y & mask ) ? 2 : 0 ) |
					( ( coordinate.z & mask ) ? 4 : 0 )
				);
			}

			// shrink coordinate depth to match tree depth
			while ( depth < coordinateDepth ) {

#if 0
				int3 const comparison = ( ( coordinate & 1 ) != 0 );
				remaining += as_float3( comparison & as_int3( remainingMaximum ) );
#else    // 0/1
				if ( coordinate.x & 1 )
					remaining.x += remainingMaximum.x;
				if ( coordinate.y & 1 )
					remaining.y += remainingMaximum.y;
				if ( coordinate.z & 1 )
					remaining.z += remainingMaximum.z;
#endif    // 0/1

				remainingMaximum *= 2;

				coordinate /= 2;
				--coordinateDepth;
			}

			// ==== descend until we reach a leaf node ====

			while ( pNode->children != OPENCL_NULL ) {

#if ( SETTING_MIPMAP != 0 )
#if 0
				float3 const mipmapDistance = ldexp( mipmapCoordinate, depth ) - ( float3 )( coordinate.x, coordinate.y, coordinate.z );
#else    // 0/1
				float const sideLengthInverse = ( 1 << depth );
				float3 const mipmapDistance = mipmapCoordinate * sideLengthInverse - ( float3 )( coordinate.x, coordinate.y, coordinate.z );
#endif    // 0/1
				if ( any( mipmapDistance >= spreadInverse ) )
					break;    // **NOTE: break
#endif    // SETTING_MIPMAP

				pNode = ( __global struct Node const* )( ( __global uchar const* )heap + pNode->children );
				childrenStack[ depth ] = pNode;
				++depth;

				coordinate *= 2;
				++coordinateDepth;

				remainingMaximum /= 2;

#if 0
				int3 const comparison = ( remaining >= remainingMaximum );
				remaining -= as_float3( comparison & as_int3( remainingMaximum ) );
				coordinate ^= ( comparison & 1 );
#else    // 0/1
				if ( remaining.x >= remainingMaximum.x ) {

					remaining.x -= remainingMaximum.x;
					coordinate.x ^= 1;
				}
				if ( remaining.y >= remainingMaximum.y ) {

					remaining.y -= remainingMaximum.y;
					coordinate.y ^= 1;
				}
				if ( remaining.z >= remainingMaximum.z ) {

					remaining.z -= remainingMaximum.z;
					coordinate.z ^= 1;
				}
#endif    // 0/1

				pNode += origin ^ (
					( ( coordinate.x & 1 ) ? 1 : 0 ) |
					( ( coordinate.y & 1 ) ? 2 : 0 ) |
					( ( coordinate.z & 1 ) ? 4 : 0 )
				);
			}

			// ==== step through the current voxel (texture) ====

			int comparisons = -1;
			if ( pNode->texture != OPENCL_NULL ) {

				comparisons = TraceTexture(    // returns -1 if drawn
					&color,
					&lambda,
					image,
					pNode->texture,
					&remaining,
					&remainingMaximum,
					origin
				);

				if ( comparisons < 0 )
					goto main_loop_end_label;    // **NOTE: goto
			}
			else {

				comparisons = (
					( ( remaining.x < remaining.y ) ? 1 : 0 ) |
					( ( remaining.y < remaining.z ) ? 2 : 0 ) |
					( ( remaining.z < remaining.x ) ? 4 : 0 )
				);
				switch( comparisons ) {

					case 0: case 7:    // any are fine--fall through
					case 1: case 3: {

						lambda += remaining.x;
						remaining.y -= remaining.x;
						remaining.z -= remaining.x;
						remaining.x = remainingMaximum.x;

						break;
					}

					case 2: case 6: {

						lambda += remaining.y;
						remaining.x -= remaining.y;
						remaining.z -= remaining.y;
						remaining.y = remainingMaximum.y;

						break;
					}

					case 4: case 5: {

						lambda += remaining.z;
						remaining.x -= remaining.z;
						remaining.y -= remaining.z;
						remaining.z = remainingMaximum.z;

						break;
					}
				}
			}

			// ==== step to the next adjacent voxel, ascending as necessary ====

			switch( comparisons ) {

				case 0: case 7:    // any are fine--fall through
				case 1: case 3: {

					if ( coordinate.x == 0 ) {    // underflow indicates that we've left the octtree

						lambda = INFINITY;
						goto main_loop_end_label;    // **NOTE: goto
					}

					--coordinate.x;
					for ( uint mask = 1; ( coordinate.x & mask ) != 0; mask *= 2, --depth );
					// **NOTE: pNode will be updated below

					break;
				}

				case 2: case 6: {

					if ( coordinate.y == 0 ) {    // underflow indicates that we've left the octtree

						lambda = INFINITY;
						goto main_loop_end_label;    // **NOTE: goto
					}

					--coordinate.y;
					for ( uint mask = 1; ( coordinate.y & mask ) != 0; mask *= 2, --depth );
					// **NOTE: pNode will be updated below

					break;
				}

				case 4: case 5: {

					if ( coordinate.z == 0 ) {    // underflow indicates that we've left the octtree

						lambda = INFINITY;
						goto main_loop_end_label;    // **NOTE: goto
					}

					--coordinate.z;
					for ( uint mask = 1; ( coordinate.z & mask ) != 0; mask *= 2, --depth );
					// **NOTE: pNode will be updated below

					break;
				}
			}

			// move pNode to the correct position at the current depth
			pNode = childrenStack[ depth - 1 ];
			uint const mask = ( 1 << ( coordinateDepth - depth ) );
			pNode += origin ^ (
				( ( coordinate.x & mask ) ? 1 : 0 ) |
				( ( coordinate.y & mask ) ? 2 : 0 ) |
				( ( coordinate.z & mask ) ? 4 : 0 )
			);
		}
main_loop_end_label:;
	}

	*colorDestination = color;
	*depthDestination = lambda;
}




/*==============================================================================
	TraceScreen function
==============================================================================*/


__kernel void TraceScreen(
	__global uint*  const colors,
	__global float* const depths,
	uint2 const dimension,
	uint const colorsPitch,
	uint const depthsPitch,
	float2 const fieldOfView,
	float const spread,
	__global void const* const heap,
	__read_only image3d_t image,
	uint const root,
	float3 const xx,
	float3 const yy,
	float3 const zz,
	float3 const pp
)
{
	uint const index = get_global_id( 0 );
	if ( index < dimension.x * dimension.y ) {

		/*
			To maximize the chances of threads in the same warp accessing the
			same data (by working on nearby rays), we work in Morton Z-order.
		*/
		uint2 const coordinate = InverseMortonZOrder( index, dimension );

		float3 const spreadX = ( 2 * fieldOfView.x / dimension.x ) * xx;
		float3 const spreadY = ( 2 * fieldOfView.y / dimension.y ) * yy;

		float3 const ray = (
			( coordinate.x + 0.5 - 0.5 * dimension.x ) * spreadX -
			( coordinate.y + 0.5 - 0.5 * dimension.y ) * spreadY -
			zz
		);

		__global uint*  const colorDestination = ( __global uint*  )( ( __global uchar* )colors + coordinate.y * colorsPitch + coordinate.x * sizeof( uint  ) );
		__global float* const depthDestination = ( __global float* )( ( __global uchar* )depths + coordinate.y * depthsPitch + coordinate.x * sizeof( float ) );
		Trace( colorDestination, depthDestination, spread, heap, image, root, pp, ray );
	}
}
