#pragma once

#ifndef WINDOWS_EXCEPTION_H
#define WINDOWS_EXCEPTION_H


/**
	\file
	\brief Contains WindowsException class.
*/




#include <winerror.h>

#include <string>
#include <stdexcept>




/*==============================================================================
	THROW_WINDOWS_EXCEPTION macro
==============================================================================*/


/// \brief Throws a WindowsException labeled with the file, line, string \p what and \c HRESULT error code \p hResult.
#define THROW_WINDOWS_EXCEPTION( what, hResult ) \
	throw ::WindowsException( __FILE__, __LINE__, ( what ), ( hResult ) )


/// \brief Evaluates \p what, and if it returns an \c HRESULT failure code, throws a WindowsException.
#define CHECK_HRESULT( what ) { \
	HRESULT result = ( what ); \
	if ( FAILED( result ) ) \
		THROW_WINDOWS_EXCEPTION( #what, result ); \
}




/*==============================================================================
	WindowsException class
==============================================================================*/


/**
	\brief Windows exception class.

	WindowsExceptions should not be constructed explicitly. Instead, use the
	THROW_WINDOWS_EXCEPTION() macro, which will construct and throw an
	exception annotated with the filename and line number of the failure.

	The difference between this class and the "default" exception class is that
	WindowsExceptions convert \c HRESULT error codes into error strings.

	The error message returned by what() has the following components:

	1. Filename and line number at which the exception was thrown.
	2. Description string \p what passed to the constructor.
	3. Description of the \c HRESULT error code passed to the constructor.
*/
struct WindowsException : public std::exception {

	WindowsException( char const* const file, unsigned int const line, std::string const& what, HRESULT const hResult );


	virtual ~WindowsException() throw();


	virtual const char* what() const throw();    ///< Returns text (UTF-8) error message for this exception.


private:

	std::string m_what;
};




#endif    // WINDOWS_EXCEPTION_H
